;; Generated package description from lex.el  -*- no-byte-compile: t -*-
(define-package "lex" "1.2" "Lexical analyser construction" 'nil :commit "c7d76551cce66901b9f172f903ce720526c0fb52" :url "https://elpa.gnu.org/packages/lex.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
